/*
 * Pointer - simple reference counted pointer.
 *
 * The is a non-intrusive implementation that allocates an additional
 * int and pointer for every counted object.
 */

/*
  Originally from http://ootips.org/yonat/4dev/counted_ptr.h and
  http://ootips.org/yonat/4dev/smart-pointers.html
  
  Modifications Copyright (c) 2008 Dave Fletcher
  All Rights Reserved
*/
 
#ifndef SP_POINTER
#define SP_POINTER

/* For ANSI-challenged compilers, you may want to #define
 * NO_MEMBER_TEMPLATES or explicit */

namespace SpinParser {

  template <class X> class Pointer {

    public:
      typedef X element_type;

      explicit Pointer(X* p = 0) : itsCounter(0) {
        if (p) itsCounter = new counter(p);
      }
      ~Pointer() { release(); }
      Pointer(const Pointer& r) throw() { acquire(r.itsCounter); }
      Pointer& operator=(const Pointer& r) {
        if (this != &r) {
          release();
          acquire(r.itsCounter);
        }
        return *this;
      }

#ifdef HAVE_MEMBER_TEMPLATES
      template <class Y> friend class Pointer<Y>;
      template <class Y> Pointer(const Pointer<Y>& r) throw() {
        acquire(r.itsCounter);
      }
      template <class Y> Pointer& operator=(const Pointer<Y>& r) {
        if (this != &r) {
          release();
          acquire(r.itsCounter);
        }
        return *this;
      }
#endif /* [HAVE_MEMBER_TEMPLATES] */

      X& operator*() const throw() { return *itsCounter->ptr;}
      X* operator->() const throw() { return itsCounter->ptr;}
      X* get() const throw() { return itsCounter ? itsCounter->ptr : 0;}
      bool unique() const throw() {
        return (itsCounter ? itsCounter->count == 1 : true);
      }

    private:
      struct counter {
        counter(X* p = 0, unsigned c = 1) : ptr(p), count(c) {}
        X* ptr;
        unsigned count;
      }* itsCounter;
      void acquire(counter* c) throw() {
          itsCounter = c;
          if (c) ++c->count;
      }
      void release() {
        if (itsCounter) {
          if (--itsCounter->count == 0) {
            delete itsCounter->ptr;
            delete itsCounter;
          }
          itsCounter = 0;
        }
      }
  };
}

#endif /* [SP_POINTER] */
